// $Id: setup_mic_ad1985.cc,v 1.5 2006/03/20 11:31:26 kostik Exp $

#include <unistd.h>
#include <sys/types.h>
#include <machine/sysarch.h>
#include <sys/pciio.h>
#include <fcntl.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

const uint16_t AC97_VND_ID1 = 0x7c;
const uint16_t AD1985_VND_ID1 = 0x4144;
const uint16_t AC97_VND_ID2 = 0x7e;
const uint16_t AD1985_VND_ID2 = 0x5375;

const uint16_t AC97_MIC_VOL = 0x0e;
const uint16_t AC97_MIC_VOL_M20 = 0x0040;

const uint16_t AD1985_SER_CONF = 0x74;
const uint16_t AD1985_SER_CONF_OMS = 0x0200;

const uint16_t AD1985_MISC_CTRL = 0x76;
const uint16_t AD1985_MISC_CTRL_CLDIS = 0x0800;

uint16_t
inw(uint16_t port)
{
	uint16_t res;

	 __asm __volatile("inw %%dx,%0" : "=a" (res) : "d" (port));
        return res;
}

void
outw(uint16_t port, uint16_t val)
{
	fprintf(stderr, "setting 0x%04x to 0x%04x\n", port, val);
  	 __asm __volatile("outw %%ax, %%dx" : : "a" (val), "d" (port));
}

void
port_set_bits(uint16_t port, uint16_t bits)
{
	uint16_t v = inw(port);
	v |= bits;
	outw(port, v);
}

void
set_bits(uint16_t base)
{
	port_set_bits(base + AC97_MIC_VOL, AC97_MIC_VOL_M20);
	port_set_bits(base + AD1985_SER_CONF, AD1985_SER_CONF_OMS);
	port_set_bits(base + AD1985_MISC_CTRL, AD1985_MISC_CTRL_CLDIS);
}

void
gain_privileges(uint16_t base, uint16_t len)
{
	int err = i386_set_ioperm(base, len, 1);
	if (err)
	{
		perror("i386_set_ioperm");
		exit(1);
	}
}

void
check_codec(uint16_t base)
{
	uint16_t id1, id2;

	id1 = inw(base + AC97_VND_ID1);
	id2 = inw(base + AC97_VND_ID2);
	if ((id1 != AD1985_VND_ID1) or (id2 != AD1985_VND_ID2))
	{
		fprintf(stderr, "Not an AD1985 codec: %04x %04x\n", id1, id2);
		exit(1);
	}
}

uint16_t
get_pcm_base(unsigned unit_num)
{
	int pcifd = open("/dev/pci", O_RDWR);
	if (pcifd == -1)
	{
		perror("opening /dev/pci");
		exit(1);
	}

	struct pci_conf_io pconf;
	struct pci_match_conf patterns[1];
	struct pci_conf pcm[1];
	memset(&pconf, 0, sizeof(pconf));
	memset(patterns, 0, sizeof(patterns));
	memset(pcm, 0, sizeof(pcm));

	pconf.pat_buf_len = sizeof(patterns);
	pconf.num_patterns = sizeof(patterns) / sizeof(patterns[0]);
	pconf.patterns = patterns;
	pconf.match_buf_len = sizeof(pcm);
	pconf.matches = pcm;

	strlcpy(patterns[0].pd_name, "pcm", sizeof(patterns[0].pd_name));
	patterns[0].pd_unit = unit_num;
	patterns[0].flags = (pci_getconf_flags)(PCI_GETCONF_MATCH_NAME | PCI_GETCONF_MATCH_UNIT);

	if (ioctl(pcifd, PCIOCGETCONF, &pconf) == -1)
	{
		perror("PCIOCGETCONF");
		exit(1);
	}
	if (pconf.num_matches < 1)
	{
		fprintf(stderr, "pcm%d not found\n", unit_num);
		exit(1);
	}

	struct pci_io pcmdev;
	memset(&pcmdev, 0, sizeof(pcmdev));
	pcmdev.pi_sel = pcm[0].pc_sel;
	pcmdev.pi_reg = 0x10;	/* NAMMBAR Native Audio Mixer Base Address */
	pcmdev.pi_width = 4;

	if (ioctl(pcifd, PCIOCREAD, &pcmdev) == -1)
	{
		perror("PCIOCREAD");
		exit(1);
	}

	uint16_t base = pcmdev.pi_data & ~1;
	if (base == 0)
	{
		fprintf(stderr, "pcm0: mixer base address is 0\n");
		exit(1);
	}

	close(pcifd);
	return base;
}

int
main(int argc, char *argv[])
{
	uint16_t base;
	unsigned unit_num;

	if (argc != 2)
	{
		fprintf(stderr, "Usage: setup_mic_ad1985 <pcm device unit>\n");
		exit(2);
	}

	unit_num = strtol(argv[1], NULL, 0);

	base = get_pcm_base(unit_num);
	gain_privileges(base, 0x80);
	check_codec(base);
	set_bits(base);

	return 0;
}
